.\" Copyright (C) 2002-2003  Ben Kibbey <bjk@arbornet.org>
.\" 
.\" This program is free software; you can redistribute it and/or modify
.\" it under the terms of the GNU General Public License as published by
.\" the Free Software Foundation; either version 2 of the License, or
.\" (at your option) any later version.
.\" 
.\" This program is distributed in the hope that it will be useful,
.\" but WITHOUT ANY WARRANTY; without even the implied warranty of
.\" MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
.\" GNU General Public License for more details.
.\" 
.\" You should have received a copy of the GNU General Public License
.\" along with this program; if not, write to the Free Software
.\" Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
.
.Dd February 05, 2003
.Os
.Dt CBOARD 6
.
.Sh NAME
.
.Nm cboard
.Nd an ncurses chess engine frontend
.
.Sh SYNOPSIS
.
.Nm cboard
.Op Fl hv
.Op Fl p Ar pgnfile
.Op Fl i Ar server Op Ar :port
.Op Fl u Ar user Op Ar :password
.
.Sh DESCRIPTION
.
.Nm cboard
is a ncurses frontend to
.Xr gnuchess .
.
.Pp
The following command line options are available:
.
.Bl -tag -width
.It Fl p Ar pgnfile
Load a PGN formatted file. If the filename extension ends with .Z, .bz, .bz2,
\&.gz or \&.zip, then it will be uncompressed (and compressed if saving)
automatically provided you have
.Xr compress 1 ,
.Xr bzip2 1 ,
.Xr gzip 1
and
.Xr zip 1
installed.
.It Fl i Ar server Op Ar :port
ICS server and option port.
.It Fl u Ar user Op Ar :password
ICS username and optional password.
.It Fl v
Version information.
.It Fl h
Program help text.
.El
.
.Sh "CONFIGURATION FILE"
Empty lines and lines beginning with a '#' are ignored. Other lines are
parameter and value pairs which must be on the same line and separated with
one or more space or TAB characters.
.Pp
.Bl -tag -compact
.It Cm book Ar string
The book method to use where string is one of: 
.Ar off , 
.Ar prefer , 
.Ar worst  ,
.Ar best
or
.Ar random . 
The default is 
.Ar random .
.Pp
.It Cm engine_cmd Ar command Op args ...
The chess engine command to execute with any arguments. The default is
\&"
.Ar gnuchess xboard
\&". Others haven\&'t been tested.
.Pp
.It Cm jumpcount Ar integer
When viewing move history, you can jump more than one move to the next or
previous move with the UP and DOWN keys. This sets the number of moves to
jump. The default is
.Ar 5 .
.Pp
.It Cm depth Ar integer
The search depth level of the engine when it's it's turn to move. The default
is
.Ar 0 .
.Pp
.It Cm agony Ar on|off
Display agony strings in the status window when a capture is made. The default
is
.Ar on .
.Pp
.It Cm historyagony Ar on|off
Display agony strings when browsing move history. If
.Cm agony
is disabled, then this will be disable also. The default is
.Ar off .
.Pp
.It Cm pgntag Ar name Op Ar value
Custom PGN tags to add to new games. This can also be used to overwrite
default tag values.
.Pp
.It Cm line_graphics Ar on|off
Enable or disable board line graphics. The default is
.Ar off .
.Pp
.It Cm save_directory Pa path
If set, the default directory where games will be saved and opened from in the
file browser. A tilde
.Pa
as the first
character will expand to your home directory.
.Pp
.It Cm save_prompt Ar on|off
When saving a game, prompt to edit PGN tag data. Or when saving in history
mode, prompt for additional save commands. The default is
.Ar on .
.Pp
.It Cm delete_prompt Ar on|off
Prompt before deleting games with the \&"D\&" command. The default is
.Ar on .
.Pp
.It Cm clevel Ar 1-9
Compression level of compressed saved files compressed with
.Xr zip 1 ,
.Xr bzip2 1
and 
.Xr gzip 1 .
.Pp
.It Cm ics_server Ar hostname
Internet chess server hostname used for network games. The default is
.Ar ics.freechess.org .
.Pp
.It Cm ics_port Ar port
ICS server port. The default is 
.Ar 5000 .
.Pp
.It Cm ics_user Ar username
Optional username for ICS authentication. The default is
.Ar \&"guest\&" .
.Pp
.It Cm ics_passwd Ar password
Password for ICS username.
.Pp
.It Cm valid_moves Ar on|off
When a piece is selected, show valid squares the piece can move to. The
.Cm board_white_moves
and
.Cm board_black_moves
color parameters can specify alternate colors for the valid move squares.
The
default is 
.Ar on .
.El
.
.Sh COLORS
.Bl -tag -compact
.It Cm board_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_selected Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_cursor Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_black Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_white Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_graphics Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_coords Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_white_moves Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm board_black_moves Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm status_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm status_title Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm status_border Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm status_notify Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm status_engine Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm tag_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm tag_title Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm tag_border Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm history_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm history_title Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm history_border Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm message_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm message_title Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm message_border Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm message_prompt Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm input_window Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm input_title Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm input_border Ar foreground Ar background Op Ar cattr Op Ar attr
.It Cm input_prompt Ar foreground Ar background Op Ar cattr Op Ar attr
.El
.Pp
Color and attributes can be modified from the configuration file. There are up
to four values for each parameter. The first two are the foreground and
background colors which must be one of:
.Ar black ,
.Ar white ,
.Ar blue ,
.Ar green ,
.Ar yellow ,
.Ar cyan ,
.Ar magenta
or
.Ar red .
.Pp
The third and fourth values are optional and set terminal attributes. The
third sets attributes for color terminals and the fourth for non-color
terminals. Multiple attributes may be assigned by separating them with a
comma. Note that not all terminals support all attributes. The following
attributes are recognized:
.Ar bold ,
.Ar reverse ,
.Ar dim ,
.Ar standout ,
.Ar underline ,
.Ar blink ,
.Ar invisible 
and
.Ar none .
.Pp
Some color parameters have a _window extension. This will set the default
color and attibutes for all characters for the given parameter set.
.
.Sh FILES
.Bl -tag -compact -width ~/.cboard/agony.data
.It Pa ~/.cboard/config
Configuration file.
.It Pa ~/.cboard/agony.data
A list of agony strings, one per line, which is displayed randomly when a
capature has been made.
.It Pa ~/.cboard/cc.data
Country codes. Used when editing the \&"Site\&" tag.
.It Pa ~/.cboard/nag.data
Appending to this file is fine for custom NAG types, but lines before 140 are
reserved by the PGN standard.
.It Pa ~/.cboard/fifo
A FIFO read by the chess engine for resuming games.
.It Pa ~/.cboard/tmpfile
Temporary file created and removed when reading compressed files.
.El
.
.Sh AUTHORS
.
.An "Ben Kibbey" Aq bjk@arbornet.org
